import stim
import numpy as np
from typing import List


class HeraldedCircuit_FREE_LD:
    def __init__(self, circuit: stim.Circuit, biased_erasure: bool, cycles: int, phys_error:float, erasure_ratio:float, ancilla_qubits:list, data_qubits:list, distance=None, loss_detection_freq=None, code=None, basis = None, SSR = True, printing=False, **kwargs) -> None:
        self.circuit = circuit
        self.biased_erasure = biased_erasure
        self.distance = distance
        self.code = code
        self.basis = basis
        self.SSR = SSR
        self.cycles = cycles
        self.printing = printing
        self.erasure_ratio = erasure_ratio
        self.phys_error = phys_error
        self.loss_detection_freq = loss_detection_freq
        self.ancilla_qubits = ancilla_qubits
        self.data_qubits = data_qubits
        self.lost_ancillas = {}  ###
        self.qec_cycles_complete = False  ###
        self.lost_ancillas_by_qec_round = {}  #!  {qec_round: [lost_ancilla_qubits]}
        self.lost_data_by_ld_round = {}  #!  {ld_round: [lost_data_qubits]}
        self.total_num_QEC_round = None
        
    def sample(self):
        pass

    
    def update_loss_lists(self, instruction, loss_detection_events, lost_data_qubits, lost_ancilla_qubits, loss_detector_ix):
        potential_lost_qubits = instruction.targets_copy()
        for q in potential_lost_qubits:
            if loss_detection_events[loss_detector_ix] == True:
                if q.value in self.data_qubits:
                    if q.value not in lost_data_qubits:
                        lost_data_qubits.append(q.value)
                        if self.printing:
                            print(f"Detected the loss of data qubit = {q.value}. Lost data qubits = {lost_data_qubits}")
                else: # q is an ancilla --> replace with a fresh noisy qubit
                    if q.value not in lost_ancilla_qubits:
                        lost_ancilla_qubits.append(q.value)
                        if self.printing:
                            print(f"Detected the loss of ancilla qubit = {q.value}. Lost ancilla qubits = {lost_ancilla_qubits}")        
            loss_detector_ix += 1
        return loss_detector_ix
        
        
    def get_loss_location(self, loss_detection_events: list):
        # Iterate through circuit. Every time we encounter a loss event (flagged by the 'I' gate), record the loss.
        loss_detector_ix = 0  # tracks the index of detectors in the circuit as we iterate.
        round_ix = 0
        inside_qec_round = False
        for instruction in self.circuit:
            if instruction.name == 'TICK': # begin of a QEC round:
                if not inside_qec_round: # beginning of round
                    if self.printing:
                        print(f"Starting QEC Round {round_ix}")
                    lost_ancilla_qubits = []
                    lost_data_qubits = []
                    qec_round = round_ix
                    ld_round = int(round_ix / self.loss_detection_freq)
                else: # end of round
                    if self.printing:
                        print(f"Finished QEC Round {round_ix}")
                    self.lost_ancillas_by_qec_round[qec_round] = lost_ancilla_qubits
                    if ld_round in self.lost_data_by_ld_round:
                        self.lost_data_by_ld_round[ld_round].extend(lost_data_qubits)
                    else:
                        self.lost_data_by_ld_round[ld_round] = lost_data_qubits
                    round_ix += 1
                
                inside_qec_round = not inside_qec_round
                continue
            
            if inside_qec_round:
                if instruction.name == 'I': # loss event
                    loss_detector_ix = self.update_loss_lists(instruction, loss_detection_events, lost_data_qubits, lost_ancilla_qubits, loss_detector_ix)
                            
            else:
                pass # we don't need to document losses outside QEC rounds because we assume there are non
                # lost_ancilla_qubits = []
                # lost_data_qubits = []
                # if instruction.name == 'I': # loss event
                #     loss_detector_ix = self.update_loss_lists(instruction, loss_detection_events, data_qubits, lost_data_qubits, lost_ancilla_qubits, loss_detector_ix)

    def heralded_new_circuit(self, loss_detection_events: list):
        """ This function takes the original circuit with places for potential losses and loss detection events, and generates 2 circuits: 1. experimental measurement circuit. 2. Theory decoding circuit. """
        # Initialization for every shot:
        self.lost_ancillas = {}  ###
        self.qec_cycles_complete = False  ###
        self.lost_ancillas_by_qec_round = {}  #!  {qec_round: [lost_ancilla_qubits]}
        self.lost_data_by_ld_round = {}  #!  {ld_round: [lost_data_qubits]}
        
        
        # First sweep: get location of lost qubits in the circuit.
        self.get_loss_location(loss_detection_events=loss_detection_events) 
        self.total_num_QEC_round = len(self.lost_ancillas_by_qec_round)
        
        if self.printing :
            print(f"lost_ancillas_by_qec_round={self.lost_ancillas_by_qec_round}")
            print(f"lost_data_by_ld_round={self.lost_data_by_ld_round}")

        # Second sweep: fill in the experimental circuit (heralded_circuit) and decoding circuit (new_lossless_circuit).
        heralded_circuit = stim.Circuit()
        new_lossless_circuit = stim.Circuit()
        loss_detector_ix = 0  # tracks the index of detectors in the circuit as we iterate.
        lost_data_qubits = []  # data_qubit_index: number_of_rounds_before_replace. if>0 --> delete entangling gates. if==0: reinitialize the qubit + depolarizing det error
        lost_ancilla_qubits = []
        round_ix = 0
        inside_qec_round = False
        LD_round = False
        for instruction in self.circuit:
            if instruction.name == 'TICK': # begin of a QEC round:
                if not inside_qec_round: # beginning of round
                    self.qec_cycles_complete = True
                    ld_round = int(round_ix / self.loss_detection_freq) # check that its the same
                    last_QEC_round = True if round_ix == self.total_num_QEC_round-1 else False
                    if (round_ix+1)%self.loss_detection_freq == 0:
                        LD_round = True
                    else:
                        LD_round = False
                    if self.printing:
                        print(f"a new QEC round number {round_ix}! is it a LD round? {LD_round} (loss detection freq = {self.loss_detection_freq})")
                else: # end of round
                    
                    # Replace lost ancilla qubits after every round.
                    if self.printing :
                        print(f"End of a QEC round! replaced ancilla qubits {lost_ancilla_qubits}")
                    lost_ancilla_qubits = [] # every QEC round we replace ancilla qubits with fresh ones
                    
                    
                    # If loss detection round (or last QEC round) --> replace data qubits with fresh noisy qubits
                    if LD_round or last_QEC_round:
                        if self.printing :
                            if LD_round:
                                print(f"We just finished a LD round! lets replace the following qubits with fresh ones: {lost_data_qubits}")
                            elif last_QEC_round:
                                print(f"We just finished the last round of QEC! now we have full detection on the loss of: {lost_data_qubits}")
                        for q in lost_data_qubits:
                            if LD_round:
                                # Heralded circuit - experiment - replacing the qubit with a fresh qubit: initialization + noise:
                                logical_qubit = self.circuit.qubit_index_to_logical_qubit(q)
                                prep = logical_qubit.get_qubit_init_basis(physical_index=q, logical_basis = self.basis)
                                heralded_circuit.append('R', [q])
                                if prep == 'RX':
                                    heralded_circuit.append('H', [q])

                            # Lossless circuit - decoding - p=1/2 channel to set w=0 (giving the decoder the loss information):
                            self.add_pauli_channel(new_lossless_circuit, [q])
                        
                        lost_data_qubits = []

                    round_ix += 1
                
                inside_qec_round = not inside_qec_round
                heralded_circuit.append('TICK')
                new_lossless_circuit.append('TICK')
                continue

            if inside_qec_round:
                loss_detector_ix = self.add_instruction(instruction, heralded_circuit, new_lossless_circuit, loss_detection_events, loss_detector_ix,
                                                        lost_data_qubits=lost_data_qubits, lost_ancilla_qubits=lost_ancilla_qubits, 
                                                        last_QEC_round=last_QEC_round, round_ix=round_ix) 
            else:
                loss_detector_ix = self.add_instruction(instruction, heralded_circuit, new_lossless_circuit, loss_detection_events, loss_detector_ix,
                                                        lost_data_qubits=lost_data_qubits, lost_ancilla_qubits=lost_ancilla_qubits,
                                                        last_QEC_round=False)
        return heralded_circuit, new_lossless_circuit
    
    def add_instruction(self, instruction, circuit: stim.Circuit, new_lossless_circuit: stim.Circuit, loss_detection_events: list, loss_detector_ix: int,
                        lost_data_qubits: list, lost_ancilla_qubits:list, last_QEC_round: bool = None, round_ix: int = None):


        # Update lost qubits lists:
        if instruction.name == 'I': # loss event
            loss_detector_ix = self.update_loss_lists(instruction, loss_detection_events, lost_data_qubits, lost_ancilla_qubits, loss_detector_ix)

        elif instruction.name in ['CZ','CX']:
            qubits = [q.value for q in instruction.targets_copy()]
            
            # Lossless circuit: the CZ is here, but we add error model to acount for errors in the decoder:
            new_lossless_circuit.append(instruction) # append the CZ gate to the lossless circuit anyway
            
            pairs = [(qubits[i], qubits[i + 1]) for i in range(0, len(qubits), 2)]
            for (c,t) in pairs:
                ancilla_target = c if (c not in self.data_qubits) else t
                data_target = c if c in self.data_qubits else t

                # Lossless circuit: Add noise to neighbor data of lost ancilla and to neighbor ancilla of lost data.
                if round_ix is not None:  #!
                    qec_round = round_ix  #!
                    ld_round = int(round_ix / self.loss_detection_freq)  #!
                    if instruction.name == 'CZ':
                        if (not self.SSR) or ((self.SSR) and (ancilla_target in self.lost_ancillas_by_qec_round[qec_round])):
                            strength = 1/4 if self.SSR else self.phys_error*(self.erasure_ratio)  # TODO: fix the 1/4 to be 1/degree_of_ancilla
                            new_lossless_circuit.append('PAULI_CHANNEL_1', [data_target], strength*np.array([0, 0, 0.5])) # {I,Z}
                            # new_lossless_circuit.append('PAULI_CHANNEL_1', [data_target], strength*np.array([0.25, 0.25, 0.25])) # {I,Z}
                        if data_target in self.lost_data_by_ld_round[ld_round]:
                            r = self.loss_detection_freq if self.cycles%self.loss_detection_freq == 0 else self.cycles%self.loss_detection_freq
                            strength = 1/(4*r) if last_QEC_round else 1/(4*self.loss_detection_freq)  # TODO: fix the 1/4 to be 1/degree_of_data. 
                            new_lossless_circuit.append('PAULI_CHANNEL_1', [ancilla_target], strength*np.array([0, 0, 0.5])) # {I,Z}  # TODO: check if the basis is right.
                            # new_lossless_circuit.append('PAULI_CHANNEL_1', [ancilla_target], strength*np.array([0.25, 0.25, 0.25])) # {I,Z}
                    elif instruction.name == 'CX':
                        if (not self.SSR) or ((self.SSR) and (ancilla_target in self.lost_ancillas_by_qec_round[qec_round])):
                            strength = 1/4 if self.SSR else self.phys_error*(self.erasure_ratio)  # TODO: fix the 1/4 to be 1/degree_of_ancilla
                            noise_channel = strength*np.array([0.5, 0, 0]) if ancilla_target == c else strength*np.array([0, 0, 0.5])
                            # new_lossless_circuit.append('PAULI_CHANNEL_1', [data_target], strength*np.array([0.5, 0, 0])) # {I,X} # TODO: causing an error!
                            new_lossless_circuit.append('PAULI_CHANNEL_1', [data_target], noise_channel)
                        if data_target in self.lost_data_by_ld_round[ld_round]:
                            r = self.loss_detection_freq if self.cycles%self.loss_detection_freq == 0 else self.cycles%self.loss_detection_freq
                            strength = 1/(4*r) if last_QEC_round else 1/(4*self.loss_detection_freq)  # TODO: fix the 1/4 to be 1/degree_of_data
                            noise_channel = strength*np.array([0.5, 0, 0]) if data_target == c else strength*np.array([0, 0, 0.5])
                            # new_lossless_circuit.append('PAULI_CHANNEL_1', [ancilla_target], strength*np.array([0.5, 0, 0])) # {I,X}  # TODO: causing an error!
                            new_lossless_circuit.append('PAULI_CHANNEL_1', [ancilla_target], noise_channel)
                
                # Heralded loss circuit: the CZ is not here is we lost one of the qubits:
                if (c in lost_data_qubits + lost_ancilla_qubits) or (t in lost_data_qubits + lost_ancilla_qubits): # remove the gate from the heralded circuit:
                    if self.printing :
                        print(f"Removing this gate from the heralded circuit: {instruction.name} {c},{t}, because my lost data qubits = {lost_data_qubits} and ancilla qubits = {lost_ancilla_qubits}")
                    pass
                else:
                    circuit.append(instruction.name, [c,t])

        elif instruction.name in ['H']:
            new_lossless_circuit.append(instruction) # append the gate to the lossless circuit anyway
            qubits = [q.value for q in instruction.targets_copy()]
            for q in qubits: # remove the gate from the heralded circuit:
                if q in lost_data_qubits + lost_ancilla_qubits:
                    if self.printing :
                        print(f"Removing this gate from the heralded circuit: {instruction.name} {q}, because my lost data qubits = {lost_data_qubits} and ancilla qubits = {lost_ancilla_qubits}")
                    pass
                else:
                    circuit.append(instruction.name, [q])

            
        elif instruction.name in ['MRX', 'MR']:
            qbts = instruction.targets_copy()
            num_measurements = len(qbts)

            # Update measurement_ixs of lost ancilla qubits.
            # for lost_ancilla_id in self.lost_ancillas:
            #     self.lost_ancillas[lost_ancilla_id] -= num_measurements

            for ix, qbt in enumerate(qbts):
                q = qbt.value

                # SSR --> ancilla loss is heralded --> create supercheck operators: ###NEW
                if q in lost_ancilla_qubits:
                    self.add_pauli_channel(new_lossless_circuit, [q])
                        
                # if q in lost_ancilla_qubits and q not in self.lost_ancillas:
                #     # store the last valid measurement ix
                #     self.lost_ancillas[q] = -(num_measurements - ix) - num_ancillas  ### ASSUME: measurement order of ancilla qubits is always the same.

                new_lossless_circuit.append(instruction.name, [q])
                
                # Heralded circuit - lost ancilla qubits give deterministic |0> measurement:
                # if SSR is on this measurement will not even be used.
                if q in lost_ancilla_qubits:
                    if instruction.name == 'MR':
                        circuit.append('R', [q])
                        circuit.append('MR', [q])
                    elif instruction.name == 'MRX':
                        circuit.append('RX', [q])
                        circuit.append('MRX', [q])
                else:
                    circuit.append(instruction.name, [q])

        elif instruction.name in ['M', 'MX']:
            qbts = instruction.targets_copy()
            num_measurements = len(qbts)
                        
            # Heralded circuit - lost qubits give deterministic |0> measurement:
            if len(lost_ancilla_qubits + lost_data_qubits) == 0:
                circuit.append(instruction)
                
            else:
                for ix, qbt in enumerate(qbts):
                    q = qbt.value
                    
                    # SSR --> ancilla loss is heralded --> create supercheck operators: ###NEW
                    if q in lost_ancilla_qubits:
                        self.add_pauli_channel(new_lossless_circuit, [q])
                    
                    if q in lost_ancilla_qubits + lost_data_qubits:
                        if instruction.name == 'M':
                            circuit.append('R', [q])
                            circuit.append('M', [q])
                        elif instruction.name == 'MX':
                            circuit.append('RX', [q])
                            circuit.append('MX', [q])
                    else:
                        circuit.append(instruction.name, [q])    
            
            new_lossless_circuit.append(instruction)
                
            # Update measurement_ixs of lost ancilla qubits.
            # for lost_ancilla_id in self.lost_ancillas:
            #     self.lost_ancillas[lost_ancilla_id] -= num_measurements
        else:
            circuit.append(instruction)
            new_lossless_circuit.append(instruction)
        return loss_detector_ix


    def add_pauli_channel(self, circuit, targets):
        if len(targets) == 1 and not self.biased_erasure:
            circuit.append('PAULI_CHANNEL_1', targets, np.array([0.25, 0.25, 0.25])) # {X,Y,Z,I}
        elif len(targets) == 1 and self.biased_erasure:
            circuit.append('PAULI_CHANNEL_1', targets, np.array([0, 0, 0.5])) # {I,Z}
        elif len(targets) == 2 and not self.biased_erasure:
            circuit.append('PAULI_CHANNEL_2', targets, [1/16 for i in range(15)]) #  {X,Y,Z,I}**2
        elif len(targets) == 2 and self.biased_erasure:
            circuit.append('PAULI_CHANNEL_2', targets, np.array([0,0,0.25,0,0,0,0,0,0,0,0,0.25,0,0,0.25])) #  biased, {Z,I}**2



import numpy as np
import qec
import time
from BiasedErasure.main_code.Simulator import *
from BiasedErasure.delayed_erasure_decoders.HeraldedCircuit_SWAP_LD import HeraldedCircuit_SWAP_LD
from BiasedErasure.main_code.noise_channels import atom_array

def Loss_MLE_Decoder_Experiment(Meta_params, distance: int, output_dir: str, measurement_events: np.ndarray, detection_events_signs: np.ndarray, use_loss_decoding=True):
        """This function decodes the loss information using mle. 
        Given heralded losses upon measurements, there are multiple potential loss events (with some probability) in the circuit.
        We use the MLE approximate decoding - for each potential loss event we create a DEM and connect all to generate the final DEM. We use MLE to decode given the ginal DEM and the experimental measurements.
        Input: Meta_params, distance, num shots, experimental data: detector shots.
        Output: final DEM, corrections, num errors.

        Meta_params = {'architecture': 'CBQC', 'code': 'Rotated_Surface', 'logical_basis': 'X', 'bias_preserving_gates': 'False', 
        'noise': 'atom_array', 'is_erasure_biased': 'False', 'LD_freq': '1', 'LD_method': 'SWAP', 'SSR': 'True', 'cycles': '2', 'ordering': 'bad', 'decoder': 'MLE',
        'circuit_type': 'memory', 'Steane_type': 'regular', 'printing': 'False', 'num_logicals': '1', 'loss_decoder': 'independent'}
        ordering: bad or fowler (good)
        decoder: MLE or MWPM
        use_loss_decoding: True or False. Do we want to use the delayed-erasure decoder?
        """
        num_shots = measurement_events.shape[0]
        
        
        # Step 0 - generate the Simulator class:
        bloch_point_params = {'erasure_ratio': '1', 'bias_ratio': '0.5'}
        # file_name = create_file_name(Meta_params, bloch_point_params = bloch_point_params)
        cycles = int(Meta_params['cycles'])
        simulator = Simulator(Meta_params=Meta_params, atom_array_sim=True, 
                                bloch_point_params=bloch_point_params, noise=atom_array , phys_err_vec=None, loss_detection_method=HeraldedCircuit_SWAP_LD, 
                                cycles = cycles, output_dir=output_dir, save_filename=None)
        
        # Step 1 - decode:
        predictions_bool, dems_list = simulator.count_logical_errors_experiment(num_shots = num_shots, distance = distance, measurement_events = measurement_events, detection_events_signs=detection_events_signs, use_loss_decoding=use_loss_decoding)
        
        
        return predictions_bool, dems_list


import stim
import numpy as np
from BiasedErasure.main_code.LogicalCircuit import LogicalCircuit
from typing import Union, List, Optional

def get_data_ancilla_indices(dx: int, dy:int):
    ancilla_qubits = [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1)]
    data_qubits =    [j*(2*dx-1)+i   for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0)]
    return data_qubits, ancilla_qubits

def data_ids(dx: int, dy: int):
    return [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0)]


def find_paired_qubit(round_type: str, qubit_index: list, dx: int, dy: int):
    # this code takes a qubit (i,j) and find the paired qubit in the SWAP round.
    i = qubit_index[0] # column
    j = qubit_index[1] # row
    
    # even rounds: up. odd rounds: down.
    paired_qubit_i = i
    if round_type == 'even':
        paired_qubit_j = j+1
    elif round_type == 'odd':
        paired_qubit_j = j-1
        if paired_qubit_j < 0:
            paired_qubit_j += 2*dy
    return [paired_qubit_i, paired_qubit_j]

def get_extra_acillas_SWAP_round(round_type: str, data_index: list, dx: int, dy: int):
    # during a SWAP round, we need to add extra ancillas such that every data qubit will have a pair.
    # even round: up. odd round: down.
    # This function return all the extra ancilla qubits and their data pairs (to do gates with them).
    ancilla_row_index = dy+1 if round_type == 'even' else -1
    new_ancillas_i_indices = range(0,2*dx-1,2)
    new_ancillas_j_indices = [2*dy-1]*dx
    pairs_list = []
    for (ancilla_i, ancilla_j) in zip(new_ancillas_i_indices, new_ancillas_j_indices):
        paired_data_i = ancilla_i
        paired_data_j = ancilla_j - 1 if round_type == 'even' else 0
        paired_data_index  = paired_data_j*(2*dx-1)+paired_data_i
        paired_ancilla_index = ancilla_j*(2*dx-1)+ancilla_i
        pairs_list.append([paired_data_index, paired_ancilla_index])
    return pairs_list
        
    
    
    
    
def ij_index_into_number(i:int, j:int, dx: int, dy: int):
    # takes a (x,y) qubit index (i,j) and return the qubit number on the lattice.
    return j*(2*dx-1)+i
    
    
def initialize(circ: LogicalCircuit, dx: int, dy: int, bias_preserving_gates: bool, basis='Z', move_duration: Optional[float] = 200):
    # Initialize data qubits
    measure_qubit_indices_odd = [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1 and i%2==1)] # odd i
    measure_qubit_indices_even = [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1 and i%2==0)] # even i
    data_qubit_indices_even = [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0 and i%2==0)] # data qubits, even i
    data_qubit_indices_odd = [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0 and i%2==1)] # data qubits, odd i
    
    if basis == 'Z':
        circ.append('MOVE_TO_NO_NOISE', measure_qubit_indices_even, 0)
        circ.append('MOVE_TO_ENTANGLING', measure_qubit_indices_odd + data_qubit_indices_even + data_qubit_indices_odd, 0)
        circ.append('RZ', sorted(measure_qubit_indices_odd + measure_qubit_indices_even + data_qubit_indices_even + data_qubit_indices_odd)) # reset all qubits to |0>
        circ.append('MOVE', measure_qubit_indices_odd + data_qubit_indices_even + data_qubit_indices_odd, move_duration) # get idling noise
        circ.append('H', sorted(data_qubit_indices_odd + measure_qubit_indices_odd)) # initialize odd ancillas + off data qubits in |+>
    elif basis == 'X':
        circ.append('MOVE_TO_NO_NOISE', measure_qubit_indices_odd, 0)
        circ.append('MOVE_TO_ENTANGLING', measure_qubit_indices_even + data_qubit_indices_even + data_qubit_indices_odd, 0)
        circ.append('RZ', sorted(measure_qubit_indices_even + measure_qubit_indices_odd + data_qubit_indices_even + data_qubit_indices_odd)) # reset all qubits to |0>
        circ.append('MOVE', measure_qubit_indices_even + data_qubit_indices_even + data_qubit_indices_odd, move_duration) # get idling noise
        circ.append('H', sorted(data_qubit_indices_even + measure_qubit_indices_even)) # initialize odd ancillas + off data qubits in |+>

    # if basis == 'Z':
    #     circ.append('MOVE_TO_NO_NOISE', measure_qubit_indices_even, 0)
    #     circ.append('MOVE_TO_ENTANGLING', measure_qubit_indices_odd + data_qubit_indices_even + data_qubit_indices_odd, 0)
    #     circ.append('RZ', measure_qubit_indices_odd + data_qubit_indices_even + data_qubit_indices_odd) # reset all data qubits + odd ancilla qubits to |0>
    #     circ.append('H', data_qubit_indices_odd + measure_qubit_indices_odd) # initialize odd ancillas + off data qubits in |+>
    # if basis == 'X':
    #     circ.append('MOVE_TO_NO_NOISE', measure_qubit_indices_odd, 0)
    #     circ.append('MOVE_TO_ENTANGLING', measure_qubit_indices_even + data_qubit_indices_even + data_qubit_indices_odd, 0)
    #     circ.append('RZ', measure_qubit_indices_even + data_qubit_indices_even + data_qubit_indices_odd) # reset all data qubits + odd ancilla qubits to |0>
    #     circ.append('H', data_qubit_indices_even + measure_qubit_indices_even) # initialize odd ancillas + off data qubits in |+>
        
    # Entangling gates
    rnd1, rnd2, rnd3, rnd4 = [], [], [], []
    for j in range(2*dy-1):
        for i in range(2*dx-1):
            if (i + j) % 2 == 1 and ((basis == 'X' and i % 2 == 0) or (basis == 'Z' and i % 2 == 1)):  # Iterate over ancilla qubits that are relevant.
                if j != 2 * dy - 2:
                    rnd1.extend([j*(2*dx-1)+i, (j+1)*(2*dx-1)+i])
                if i != 2 * dx - 2:
                    rnd2.extend([j*(2*dx-1)+i, j*(2*dx-1)+(i+1)])
                if i != 0:
                    rnd3.extend([j*(2*dx-1)+i, j*(2*dx-1)+(i - 1)])
                if j != 0:
                    rnd4.extend([j*(2*dx-1)+i, (j-1)*(2*dx-1)+i])
    
    # round 1 - CZ: (no movement errors in first round, because we counted to movement when we moved the qubits to entangling zone)
    circ.append('CZ', rnd1)
    # circ.append('TICK')
    
    # round 2 - CX:
    if bias_preserving_gates:
        circ.append('CX', rnd2)
    else:
        targets = rnd2[1::2]
        circ.append('H', targets)
        circ.append('CZ', rnd2)
        circ.append('H', targets)
    circ.append('MOVE', (), move_duration) # idle errors during movement
    # circ.append('TICK')
    
    # round 3 - CX:
    if bias_preserving_gates:
        circ.append('CX', rnd3)
    else:
        targets = rnd3[1::2]
        circ.append('H', targets)
        circ.append('CZ', rnd3)
        circ.append('H', targets)
    circ.append('MOVE', (), move_duration) # idle errors during movement
    # circ.append('TICK')
    
    # round 4 - CZ:
    circ.append('CZ', rnd4)
    circ.append('MOVE', (), move_duration) # idle errors during movement
    # circ.append('TICK')
    
    
    
    # measure checks:
    m = int(0.5*(2*dx-1)*(2*dy-1))
    if basis == 'Z':
        circ.append('H', measure_qubit_indices_odd)
        circ.append('MOVE', measure_qubit_indices_odd, move_duration) # Put idling noise on all active qubits due to movement time
        circ.append('M', sorted(measure_qubit_indices_odd + measure_qubit_indices_even)) # measure all ancilla qubits
        circ.append('MOVE_TO_NO_NOISE', measure_qubit_indices_odd + measure_qubit_indices_even, 0) # all measured qubits are inactive now
        check_ixs = [int((j*(2*dx-1)+i)/2) for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1 and i%2==0)] # deterministic checks
        for check_ix in check_ixs:
            circ.append('DETECTOR', [stim.target_rec(-(m-check_ix))])
    else:
        circ.append('H', measure_qubit_indices_even)
        circ.append('MOVE', measure_qubit_indices_even, move_duration) # Put idling noise on all active qubits due to movement time
        circ.append('M', sorted(measure_qubit_indices_odd + measure_qubit_indices_even)) # measure all ancilla qubits
        circ.append('MOVE_TO_NO_NOISE', measure_qubit_indices_odd + measure_qubit_indices_even, 0) # all measured qubits are inactive now
        check_ixs = [int((j*(2*dx-1)+i)/2) for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1 and i%2==1)] # deterministic checks
        for check_ix in check_ixs:
            circ.append('DETECTOR', [stim.target_rec(-(m-check_ix))])
                
                
def measure_stabilizers(circ: LogicalCircuit, dx: int, dy: int, bias_preserving_gates: bool, move_duration: Optional[float] = 200, offset : int = 0):
    circ.offset = 0
    measure_qubit_indices_odd = [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1 and i%2==1)]
    measure_qubit_indices_even = [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1 and i%2==0)]
    data_qubit_indices_even = [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0 and i%2==0)] # data qubits, even i
    data_qubit_indices_odd = [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0 and i%2==1)] # data qubits, odd i
    
    circ.append('MOVE_TO_ENTANGLING', data_qubit_indices_even + data_qubit_indices_odd, 0)
    circ.append('MOVE_TO_ENTANGLING', measure_qubit_indices_odd + measure_qubit_indices_even, 0)
    circ.append('MOVE', measure_qubit_indices_odd + measure_qubit_indices_even, move_duration) # get idling errors
    circ.append('RZ', sorted(measure_qubit_indices_even + measure_qubit_indices_odd)) # reset all ancilla qubits in |0>
    circ.append('H', sorted(measure_qubit_indices_even + measure_qubit_indices_odd)) # initialize all ancilla qubits in |+>
    
    # Entangling gates
    rnd1, rnd2, rnd3, rnd4 = [], [], [], []
    for j in range(2*dy-1):
        for i in range(2*dx-1):
            if (i + j) % 2 == 1:  # Iterate over all ancilla qubits.
                if j != 2 * dy - 2:
                    rnd1.extend([j*(2*dx-1)+i, (j+1)*(2*dx-1)+i])
                if i != 2 * dx - 2:
                    rnd2.extend([j*(2*dx-1)+i, j*(2*dx-1)+(i+1)])
                if i != 0:
                    rnd3.extend([j*(2*dx-1)+i, j*(2*dx-1)+(i - 1)])
                if j != 0:
                    rnd4.extend([j*(2*dx-1)+i, (j-1)*(2*dx-1)+i])
    
    # round 1 - CZ: (no movement errors in first round, because we counted to movement when we moved the qubits to entangling zone)
    circ.append('CZ', rnd1)
    # circ.append('TICK')
    
    # round 2 - CX:
    if bias_preserving_gates:
        circ.append('CX', rnd2)
    else:
        targets = rnd2[1::2]
        circ.append('H', targets)
        circ.append('CZ', rnd2)
        circ.append('H', targets)
    circ.append('MOVE', (), move_duration)
    # circ.append('TICK')
    
    # round 3 - CX:
    if bias_preserving_gates:
        circ.append('CX', rnd3)
    else:
        targets = rnd3[1::2]
        circ.append('H', targets)
        circ.append('CZ', rnd3)
        circ.append('H', targets)
    circ.append('MOVE', (), move_duration)
    # circ.append('TICK')
    
    # round 4 - CZ:
    circ.append('CZ', rnd4)
    circ.append('MOVE', (), move_duration)
    # circ.append('TICK')
    
    
    measure_qubit_indices_odd = [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1 and i%2==1)] # ancilla qubits, odd i
    measure_qubit_indices_even = [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1 and i%2==0)] # ancilla qubits, even i
    m = int(0.5*(2*dx-1)*(2*dy-1))
    circ.append('H', measure_qubit_indices_odd + measure_qubit_indices_even)
    circ.append('MOVE', measure_qubit_indices_odd + measure_qubit_indices_even, move_duration) # Put idling noise due to movement time
    circ.append('M', sorted(measure_qubit_indices_odd + measure_qubit_indices_even))
    circ.append('MOVE_TO_NO_NOISE', measure_qubit_indices_odd + measure_qubit_indices_even, 0) # all measured qubits are inactive now
    
    # Compare parity of check at t (now) vs t-1.
    # Add offset equal to number of entangling gates. One detector per gate to deal with erasure.
    for i in range(m):
        circ.append('DETECTOR', [stim.target_rec(-(i+1)),
                                stim.target_rec(-(i+1+m+circ.offset))])


def repeat_block(dx: int, dy: int, bias_preserving_gates : bool, error_model: dict, offset=0, **kwargs) -> stim.Circuit:
    data_qubits, ancilla_qubits = get_data_ancilla_indices(dx,dy)
    qubit_indices = data_qubits + ancilla_qubits
    circ2 = LogicalCircuit(qubit_indices = qubit_indices, **error_model)
    
    # circ2 = LogicalCircuit(qubit_indices = qubit_indices, initialize_circuit=False,
    #                         idle_loss_rate=error_model["idle_loss_rate"],
    #                         idle_error_rate=error_model["idle_error_rate"],
    #                         entangling_error_rate=error_model["entangling_error_rate"],
    #                         entangling_loss_rate=error_model["entangling_loss_rate"],
    #                         reset_error_rate=error_model["reset_error_rate"], measurement_error_rate=error_model["measurement_error_rate"],
    #                         single_qubit_error_rate=error_model["single_qubit_error_rate"],
    #                         reset_loss_rate=error_model["reset_loss_rate"])
    circ2.append('TICK')
    measure_stabilizers(circ2, dx, dy, offset=offset, bias_preserving_gates=bias_preserving_gates)
    return circ2


def measure_data(circ: LogicalCircuit, dx: int, dy:int, basis='Z', move_duration: Optional[float] = 200):
    # last round, measure data qubits and make them not active anymore
    # measure_qubit_indices_odd = [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1 and i%2==1)]
    # measure_qubit_indices_even = [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1 and i%2==0)]
    circ.offset = 0
    data_qubit_indices_even = [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0 and i%2==0)] # data qubits, even i
    data_qubit_indices_odd = [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0 and i%2==1)] # data qubits, odd i
    
    n = int(np.ceil(0.5*(2*dx-1)*(2*dy-1)))  # num data
    m = int(0.5*(2*dx-1)*(2*dy-1))  # num ancilla
    circ.append('MOVE_TO_ENTANGLING', data_qubit_indices_even + data_qubit_indices_odd, 0) # make sure they are in the entangling zone for the gate and measurement errors.
    # circ.append('MOVE', data_qubit_indices_even + data_qubit_indices_odd, move_duration) # no idling noise because we take the picture in the entangling zone.
    # Measure data
    # (meas1, meas2) = ('MZ', 'MX') if basis == 'Z' else ('MX', 'MZ')
    if basis == 'Z':
        for j in range(2*dy-1):
            for i in range(2*dx-1):
                if (i+j)%2 == 0: # data qubit
                    if i%2==0: # even data
                        circ.append('MZ', [j*(2*dx-1)+i])
                    else: # odd data
                        circ.append('H', [j*(2*dx-1)+i])
                        circ.append('MZ', [j*(2*dx-1)+i])
    elif basis == 'X':
        for j in range(2*dy-1):
            for i in range(2*dx-1):
                if (i+j)%2 == 0: # data qubit
                    if i%2==0: # even data
                        circ.append('H', [j*(2*dx-1)+i])
                        circ.append('MZ', [j*(2*dx-1)+i])
                    else: # odd data
                        circ.append('MZ', [j*(2*dx-1)+i])

    circ.append('MOVE_TO_NO_NOISE', data_qubit_indices_even + data_qubit_indices_odd, 0) # all measured qubits are inactive now
    
    # take care of detectors:
    anc_ix = 0
    for j in range(2*dy-1):
        for i in range(2*dx-1):
            if (i + j) % 2 == 1:
                if (basis == 'Z' and i % 2 == 0) or (basis == 'X' and i % 2 == 1):
                    detector_targets = [stim.target_rec(-(n+m-anc_ix+circ.offset))] # ancilla measurement from previous round
                    # now we will compare it to data measurements in the final round (neighbors of the ancilla)
                    if j != 2 * dy - 2:
                        data_ix = int(((j+1)*(2*dx-1)+i)/2)  # up
                        detector_targets.append(stim.target_rec(-(n-data_ix)))
                    if i != 2 * dx - 2:
                        data_ix = int((j*(2*dx-1)+(i+1))/2)  # right
                        detector_targets.append(stim.target_rec(-(n-data_ix)))
                    if i != 0:
                        data_ix = int((j*(2*dx-1)+(i - 1))/2)  # left
                        detector_targets.append(stim.target_rec(-(n-data_ix)))
                    if j != 0:
                        data_ix = int(((j-1)*(2*dx-1)+i)/2)  # down
                        detector_targets.append(stim.target_rec(-(n-data_ix)))
                    circ.append('DETECTOR', detector_targets)
                anc_ix += 1


def logical(circ: stim.Circuit, dx: int, dy: int, basis='Z') -> None:
    n = int(np.ceil(0.5*(2*dx-1)*(2*dy-1)))  # num data
    detector_targets = []

    if basis == 'Z':
        for i in range(2*dx-1):
            if i % 2 == 0:  # data qubit
                data_ix = int(i / 2)
                detector_targets.append(stim.target_rec(-(n-data_ix)))
        circ.append('OBSERVABLE_INCLUDE', detector_targets, 0)
    else:
        assert basis == 'X'
        for j in range(2*dy - 1):
            if j % 2 == 0:  # data qubit
                data_ix = int(j*(2*dx-1)/2)
                detector_targets.append(stim.target_rec(-(n-data_ix)))
        circ.append('OBSERVABLE_INCLUDE', detector_targets, 0)


def xzzx_circuit_noisy(cycles, dx: int, dy: int, bias_preserving_gates:bool, basis='Z', **kwargs):

    error_model = {
        "entangling_zone_error_rate": (.002 / 4, .002 / 4, .005 / 4),
        "entangling_gate_error_rate": (.002 / 4, .002 / 4, .0025 / 4, .002 / 4, 0, 0, 0, .002 / 4, 0, 0, 0, .0025 / 4, 0, 0, .005 / 4),
        "idle_loss_rate": 1e-7,
        "idle_error_rate": (5e-6 / 25, 5e-6 / 25, 2e-5 / 25),
        "entangling_loss_rate": 0.005/4,
        "reset_error_rate": 0.003,
        "measurement_error_rate": 0.004,
        "single_qubit_error_rate": (1e-4, 1e-4, 1e-4),
        "reset_loss_rate": 0.000
    }
    # entangling_error_rate_yes_gate = (.002 / 4, .002 / 4, .0025 / 4, .002 / 4, 0, 0, 0, .002 / 4, 0, 0, 0, .0025 / 4, 0, 0, .005 / 4)
    move_duration = 200
    data_qubits, ancilla_qubits = get_data_ancilla_indices(dx,dy)
    qubit_indices = data_qubits + ancilla_qubits
    circ = LogicalCircuit(qubit_indices = qubit_indices, initialize_circuit=False, **error_model)
    
    # circ = stim.Circuit()
    offset = 0
    initialize(circ, dx, dy, basis=basis, bias_preserving_gates=bias_preserving_gates, move_duration=move_duration)
    repeat_circ = repeat_block(dx, dy, bias_preserving_gates=bias_preserving_gates, error_model=error_model, **kwargs) * (cycles - 1)
    circ += repeat_circ
    measure_data(circ, dx, dy, basis=basis, move_duration=move_duration)
    logical(circ, dx, dy, basis=basis)
    return circ


if __name__ == '__main__':
    dx = 3
    dy = 3
    # circuit = xzzx_circuit(max(dx, dy), dx, dy, basis='X', noise=biased_erasure_noise, p2q=0.1, bias=0.5, erasure_weight=0.2, biased_erasure=False, bias_preserving=True)
    # circuit = xzzx_circuit(max(dx, dy), dx, dy, basis='Z', noise=no_noise)
    # dem = circuit.detector_error_model(allow_gauge_detectors=True,decompose_errors=True, approximate_disjoint_errors=True)
    
    
    circuit = xzzx_circuit_noisy(cycles=max(dx,dy), dx=dx, dy=dy, basis='Z', bias_preserving_gates = True)
    print(circuit)
    dem = circuit.detector_error_model(allow_gauge_detectors=True,decompose_errors=True, approximate_disjoint_errors=True, ignore_decomposition_failures=True)
    #circuit.diagram("timeline-svg")
    # graph = pymatching.Matching.from_detector_error_model(dem)
    # graph = pymatching.Matching.to_networkx(graph)
    # draw_graph(graph)
    #print(dem)
    print(repr(dem))
    #dem.diagram("matchgraph-3d")
    
    # circuit = xzzx_circuit(max(dx, dy), dx, dy, noise=two_q_gate_err, p2q=0.2)
    # circuit = biased_erasure_noise(circuit, p2q=0.01, bias=1, erasure_weight=0, biased_erasure=False, bias_preserving=True)
    


from . import Experimental_Loss_Decoder
from .HeraldedCircuit_SWAP_LD import HeraldedCircuit_SWAP_LD
from .MLE_Loss_Decoder import MLE_Loss_Decoder
from .HeraldedCircuit_FREE_LD import HeraldedCircuit_FREE_LD


import stim
import numpy as np
import random



def biased_erasure_noise_old(phys_err, bias_ratio, erasure_ratio):
    # here we put both loss and Pauli errors on each qubit.
    phys_err_1q = (1- np.sqrt(1 - phys_err))
    px_q1 = phys_err_1q / (2*(1+bias_ratio))
    py_q1 = px_q1
    pz_q1 = bias_ratio*(px_q1 + py_q1)
    entangling_gate_error_rate=(1-erasure_ratio)*np.array([px_q1, py_q1, pz_q1])
    entangling_gate_loss_rate=erasure_ratio*phys_err_1q
    return entangling_gate_error_rate, entangling_gate_loss_rate


def biased_erasure_noise(phys_err, bias_ratio):
    # here entangling_gate_loss_rate, entangling_gate_error_rate don't depend on erasure_ratio because this will be counted inside LogicalCircuit
    phys_err_1q = 1- np.sqrt(1 - phys_err)
    px_q1 = phys_err_1q / (2*(1+bias_ratio))
    py_q1 = px_q1
    pz_q1 = bias_ratio*(px_q1 + py_q1)
    entangling_gate_error_rate=np.array([px_q1, py_q1, pz_q1])
    entangling_gate_loss_rate=phys_err_1q
    return entangling_gate_error_rate, entangling_gate_loss_rate


def biased_erasure_noise_MBQC(
        circuit: stim.Circuit, dx: int = 2, dy: int = 2, 
        entangling_gate_error_rate = np.array([0.0,0.0,0.0]),
        entangling_gate_loss_rate = 0.0,
        erasure_weight: float = 0.0,
        bias_preserving: bool = False,
         **kwargs
        ) -> stim.Circuit:
    

    result = stim.Circuit()
    
    # Handling loss:
    potential_lost_qubits = np.array([], dtype=np.int32)
    loss_probabilities = np.array([], dtype=np.float32)
    
    n = int(np.ceil(0.5*(2*dx-1)*(2*dy-1)))  # num data qubit in a layer
    m = int((2*dx-1)*(2*dy-1)*0.5)  # num ancilla qubits in a surface code (twice num ancilla in MBQC layer)
    single_layer_offset = n + m
    
    for instruction in circuit:
        if (instruction.name in ['CZ', 'CX']):
            qubits = instruction.targets_copy()
            layer_indices = [qbt.value // single_layer_offset for qbt in qubits]
            if set(layer_indices).issubset([0]):  # no error on gates within the first layer
                result.append(instruction)
            else:
                loss_probabilities, potential_lost_qubits = entangling_error(instruction, result, entangling_gate_error_rate, 
                                entangling_gate_loss_rate, erasure_weight, bias_preserving, 
                                loss_probabilities, potential_lost_qubits)

        else:
            result.append(instruction)
    return result, loss_probabilities, potential_lost_qubits



def entangling_error(instruction, circuit: stim.Circuit, entangling_gate_error_rate, entangling_gate_loss_rate, erasure_weight,
                    bias_preserving: bool, loss_probabilities: np.array, potential_lost_qubits: np.array):
    # targets = instruction.targets_copy()
    
    targets = [t.value for t in instruction.targets_copy()]
    
    ### Part 1 - bias preserving gates:
    if instruction.name == 'CX' and not bias_preserving:
        h_targets = targets[1::2]
        circuit.append('H', h_targets)
        circuit.append('CZ', targets)
        
    else:
        circuit.append(instruction)
        

    ### Part 2 - Put noise:
    err_types = random.choices(['loss','pauli'], weights=[erasure_weight,(1-erasure_weight)], k=int(len(targets)))
    
    # Pauli noise channel:
    pauli_targets = [target for target, err_type in zip(targets, err_types) if err_type == 'pauli'] # take all targets with 'pauli' errors:
    pauli_error = np.array(entangling_gate_error_rate)
    if sum(pauli_error) > 0 and len(pauli_targets)> 0:
        if len(pauli_error) == 3:
            circuit.append('PAULI_CHANNEL_1', pauli_targets, pauli_error)
        elif len(pauli_error) == 8:
            circuit.append('PAULI_CHANNEL_2', pauli_targets, pauli_error)
    
    loss_prob = entangling_gate_loss_rate
    if loss_prob>0:
        # Apply I channel to all targets to simulate identity operation
        circuit.append('I', targets)

        # Update the potential lost qubits list and loss probabilities
        for target, err_type in zip(targets, err_types):
            if err_type == 'loss':
                potential_lost_qubits = np.append(potential_lost_qubits, target)
                loss_probabilities = np.append(loss_probabilities, loss_prob)
            else:
                # If it's not a loss, append zero to maintain the correct indices
                potential_lost_qubits = np.append(potential_lost_qubits, target)
                loss_probabilities = np.append(loss_probabilities, 0)



    ### Part 3 - bias preserving gates:
    if instruction.name == 'CX' and not bias_preserving:
        circuit.append('H', h_targets)

    
    return loss_probabilities, potential_lost_qubits
            

def biased_erasure_noise_correlated(phys_err, bias_ratio):
    print("Need to build this feature")
    return None

def no_noise(phys_err, bias_ratio):
    entangling_gate_error_rate = (0,0,0)
    entangling_gate_loss_rate = 0
    return entangling_gate_error_rate, entangling_gate_loss_rate


def atom_array(phys_err, bias_ratio):
    return None, None



if __name__ == '__main__':
    pass




from BiasedErasure.main_code.simulate import *
import numpy as np
from distutils.util import strtobool



if __name__ == '__main__':
    """_summary_
    This code is here to collect data on the bloch sphere. Run it many times to get enough data (at least 1000 jobs).
    An example of an input in the .sh file:
    "python build_bloch_sphere.py --architecture=CBQC --code=XZZX --logical_basis=X --bias_preserving_gates=True 
    --noise=biased_erasure_noise --with_erasures=True --is_erasure_biased=False 
    --output_dir=/Users/gefenbaranes/Documents/results --num_shots=5 --distances=9,11,13 --num_p_phys=15 --loss_detection_freq=2 --SRR=False --cycles=d"
    """
    
    Meta_params_keys = ['architecture', 'code', 'logical_basis', 'bias_preserving_gates', 
                'noise', 'is_erasure_biased', 'LD_freq', 'LD_method', 'SSR', 'cycles', 'ordering', 'decoder', 'circuit_type',
                'printing', 'num_logicals', 'Steane_type', 'loss_decoder']
    Meta_params = {key: None for key in Meta_params_keys}
    
    distances=[5,7,9]; num_shots=30 # default values, user will change them
    # we take a small number because we loop here over many phi and theta. One needs to run this function over many jobs to collect enough data.
    
    for k, v in ((k.lstrip('-'), v) for k,v in (a.split('=') for a in sys.argv[1:])):
        if k == 'num_p_phys':
            Meta_params[k] = int(v)
        elif k in Meta_params:
            Meta_params[k] = v
        elif k =='output_dir':
            output_dir = v
        elif k == 'distances':
            distances= [int(x) for x in v.split(',')]
        elif k == 'num_shots':
            num_shots = int(v)
    print(f"Meta params dictionary = {Meta_params}")
    
    # theta_vec = np.linspace(0, np.pi/2, 7)
    # phi_vec = np.linspace(0, np.pi/2, 7)
    # thresholds_mat = np.ndarray((len(theta_vec),len(phi_vec)), dtype = 'float')
    # erasure_ratio_vec = np.concatenate([np.linspace(0,1,11),[0.98]]); erasure_ratio_vec.sort()
    # bias_ratio_vec = np.linspace(0,1,11)
    # erasure_ratio_vec = [1.0, 0.75, 0.5, 0.25, 0.0]
    # bias_ratio_vec = [0.5, 10, 50, 100]
    erasure_ratio_vec = [1.0]
    bias_ratio_vec = [0.5]
    
    for i in range(len(erasure_ratio_vec)):
        for j in range(len(bias_ratio_vec)):
            erasure_ratio = erasure_ratio_vec[i]
            bias_ratio = bias_ratio_vec[j]
            # erasure_ratio = (np.cos(phi))**2
            # bias_ratio = (np.cos(theta))**2
            bloch_point_params = {'erasure_ratio': erasure_ratio, 'bias_ratio': bias_ratio}
            print(f"bloch point params dictionary = {bloch_point_params}")
            # Estimating the threshold to know which range of phys_error we should run:
            num_p_phys = Meta_params['num_p_phys']
            # cond_0 = strtobool(Meta_params["with_erasures"])
            cond_1 = strtobool(Meta_params['is_erasure_biased'])
            cond_2 = strtobool(Meta_params["bias_preserving_gates"])
            phys_err_vec = np.round(np.linspace(3.5e-2, 5.5e-2, num_p_phys),8)
                    
            Bloch_sphere_point_simulate(Meta_params=Meta_params, bloch_point_params=bloch_point_params, output_dir=output_dir,
                                        phys_err_vec=phys_err_vec, distances=distances, num_shots=num_shots)
            # file_name = create_tile_name(Meta_params, bloch_point_params)
            # threshold = calculate_logical_prob(folder=output_dir, filename=file_name, title=file_name, Erasure_ratio=erasure_ratio, 
            #                                 Bias=bias_ratio, plotting = False, distances=distances, num_shots=num_shots)
            # threshold[i,j] = threshold
        
    



from . import circuits
from . import LogicalCircuit
from . import LogicalCircuitMBQC
from . import noise_channels
from . import simulate
from . import Simulator
from . import StabilizerCode
from . import utilities
from . import xzzx_circuit_noisy
from . import xzzx_circuit
from . import xzzx_cluster
from . import xzzx_cluster_noisy
from . import XZZX




import stim
import time
from BiasedErasure.main_code.utilities import draw_graph
import numpy as np
from BiasedErasure.main_code.noise_channels import biased_erasure_noise

def get_data_ancilla_indices(dx: int, dy:int):
    ancilla_qubits = [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1)]
    data_qubits =    [j*(2*dx-1)+i   for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0)]
    return data_qubits, ancilla_qubits

def data_ids(dx: int, dy: int):
    return [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0)]

def ancilla_ids(dx: int, dy: int):
    return [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1)]

def find_paired_qubit(round_type: str, qubit_index: list, dx: int, dy: int):
    # this code takes a qubit (i,j) and find the paired qubit in the SWAP round.
    i = qubit_index[0] # column
    j = qubit_index[1] # row
    
    # even rounds: up. odd rounds: down.
    paired_qubit_i = i
    if round_type == 'even':
        paired_qubit_j = j+1
    elif round_type == 'odd':
        paired_qubit_j = j-1
        if paired_qubit_j < 0:
            paired_qubit_j += 2*dy
    return [paired_qubit_i, paired_qubit_j]

def get_extra_acillas_SWAP_round(round_type: str, data_index: list, dx: int, dy: int):
    # during a SWAP round, we need to add extra ancillas such that every data qubit will have a pair.
    # even round: up. odd round: down.
    # This function return all the extra ancilla qubits and their data pairs (to do gates with them).
    ancilla_row_index = dy+1 if round_type == 'even' else -1
    new_ancillas_i_indices = range(0,2*dx-1,2)
    new_ancillas_j_indices = [2*dy-1]*dx
    pairs_list = []
    for (ancilla_i, ancilla_j) in zip(new_ancillas_i_indices, new_ancillas_j_indices):
        paired_data_i = ancilla_i
        paired_data_j = ancilla_j - 1 if round_type == 'even' else 0
        paired_data_index  = paired_data_j*(2*dx-1)+paired_data_i
        paired_ancilla_index = ancilla_j*(2*dx-1)+ancilla_i
        pairs_list.append([paired_data_index, paired_ancilla_index])
    return pairs_list
        
    
    
    
    
def ij_index_into_number(i:int, j:int, dx: int, dy: int):
    # takes a (x,y) qubit index (i,j) and return the qubit number on the lattice.
    return j*(2*dx-1)+i
    
    
def initialize(circ: stim.Circuit, dx: int, dy: int, basis='Z'):
    # Initialize data qubits
    (prep1, prep2) = ('RZ', 'RX') if basis == 'X' else ('RX', 'RZ')
    circ.append(prep1, [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0 and i%2==0)]) # data qubits, even i
    circ.append(prep2, [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0 and i%2==1)]) # data qubits, odd i

    # Initialize ancilla qubits
    circ.append('RX', [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if (i+j)%2 == 1])
    circ.append('TICK')


def checks(circ: stim.Circuit, dx: int, dy: int):
    # Entangling gates
    rnd1, rnd2, rnd3, rnd4 = [], [], [], []
    for j in range(2*dy-1):
        for i in range(2*dx-1):
            if (i + j) % 2 == 1:  # Iterate over ancilla qubits.
                if j != 2 * dy - 2:
                    rnd1.extend([j*(2*dx-1)+i, (j+1)*(2*dx-1)+i])
                if i != 2 * dx - 2:
                    rnd2.extend([j*(2*dx-1)+i, j*(2*dx-1)+(i+1)])
                if i != 0:
                    rnd3.extend([j*(2*dx-1)+i, j*(2*dx-1)+(i - 1)])
                if j != 0:
                    rnd4.extend([j*(2*dx-1)+i, (j-1)*(2*dx-1)+i])
    circ.append('CZ', rnd1)
    circ.append('TICK')
    circ.append('CX', rnd2)
    circ.append('TICK')
    circ.append('CX', rnd3)
    circ.append('TICK')
    circ.append('CZ', rnd4)
    circ.append('TICK')


def measure_checks(circ, dx: int, dy: int, t_compare=False, basis='Z', offset=0):
    m = int(0.5*(2*dx-1)*(2*dy-1))
    circ.append('MRX', [j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if (i+j)%2 == 1])
    if t_compare:
        # Compare parity of check at t (now) vs t-1.

        # Add offset equal to number of entangling gates. One detector per gate to deal with erasure.
        # offset = 2*(dx-1)*(2*dy-1) + 2*(dy-1)*(2*dx-1) # num of entangling gates. For erasure after every entangling gate
        # offset = 0 # for erasure only after initialization
        for i in range(m):
            circ.append('DETECTOR', [stim.target_rec(-(i+1)),
                                    stim.target_rec(-(i+1+m+offset))])
    else:
        if basis == 'X':
            check_ixs = [int((j*(2*dx-1)+i)/2) for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1 and i%2==0)]
            for check_ix in check_ixs:
                circ.append('DETECTOR', [stim.target_rec(-(m-check_ix))])
        else:
            check_ixs = [int((j*(2*dx-1)+i)/2) for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1 and i%2==1)]
            for check_ix in check_ixs:
                circ.append('DETECTOR', [stim.target_rec(-(m-check_ix))])


def repeat_block(dx: int, dy: int, offset=0, **kwargs) -> stim.Circuit:
    circ2 = stim.Circuit()
    circ2.append('TICK')
    checks(circ2, dx, dy)
    measure_checks(circ2, dx, dy, t_compare=True, offset=offset)
    return circ2


def measure_data(circ: stim.Circuit, dx: int, dy:int, basis='Z'):
    n = int(np.ceil(0.5*(2*dx-1)*(2*dy-1)))  # num data
    m = int(0.5*(2*dx-1)*(2*dy-1))  # num ancilla
    # Measure data
    (meas1, meas2) = ('MZ', 'MX') if basis == 'X' else ('MX', 'MZ')
    for j in range(2*dy-1):
        for i in range(2*dx-1):
            if (i+j)%2 == 0:
                if i%2==0:
                    circ.append(meas1, [j*(2*dx-1)+i])
                else:
                    circ.append(meas2, [j*(2*dx-1)+i])

    anc_ix = 0
    for j in range(2*dy-1):
        for i in range(2*dx-1):
            if (i + j) % 2 == 1:
                if (basis == 'X' and i % 2 == 0) or (basis == 'Z' and i % 2 == 1):
                    detector_targets = [stim.target_rec(-(n+m-anc_ix))] # ancilla measurement from previous round
                    # now we will compare it to data measurements in the final round (neighbors of the ancilla)
                    if j != 2 * dy - 2:
                        data_ix = int(((j+1)*(2*dx-1)+i)/2)  # up
                        detector_targets.append(stim.target_rec(-(n-data_ix)))
                    if i != 2 * dx - 2:
                        data_ix = int((j*(2*dx-1)+(i+1))/2)  # right
                        detector_targets.append(stim.target_rec(-(n-data_ix)))
                    if i != 0:
                        data_ix = int((j*(2*dx-1)+(i - 1))/2)  # left
                        detector_targets.append(stim.target_rec(-(n-data_ix)))
                    if j != 0:
                        data_ix = int(((j-1)*(2*dx-1)+i)/2)  # down
                        detector_targets.append(stim.target_rec(-(n-data_ix)))
                    circ.append('DETECTOR', detector_targets)
                anc_ix += 1


def logical(circ: stim.Circuit, dx: int, dy: int, basis='Z') -> None:
    n = int(np.ceil(0.5*(2*dx-1)*(2*dy-1)))  # num data
    detector_targets = []

    if basis == 'X':
        for i in range(2*dx-1):
            if i % 2 == 0:  # data qubit
                data_ix = int(i / 2)
                detector_targets.append(stim.target_rec(-(n-data_ix)))
        circ.append('OBSERVABLE_INCLUDE', detector_targets, 0)
    else:
        assert basis == 'Z'
        for j in range(2*dy - 1):
            if j % 2 == 0:  # data qubit
                data_ix = int(j*(2*dx-1)/2)
                detector_targets.append(stim.target_rec(-(n-data_ix)))
        circ.append('OBSERVABLE_INCLUDE', detector_targets, 0)


def xzzx_circuit(cycles, dx: int, dy: int, basis='Z', **kwargs):
    circ = stim.Circuit()
    initialize(circ, dx, dy, basis=basis)
    checks(circ, dx, dy) # note that we do all gates in the preparation round and don't add noise. If you want to add noise to prep rounds, do only half of the gates!
    measure_checks(circ, dx, dy, basis=basis)
    repeat_circ = repeat_block(dx, dy, **kwargs) * (cycles - 1)
    # repeat_circ = noise(repeat_circ, **kwargs)
    circ += repeat_circ
    measure_data(circ, dx, dy, basis=basis)
    logical(circ, dx, dy, basis=basis)
    return circ


if __name__ == '__main__':
    dx = 3
    dy = 3
    # circuit = xzzx_circuit(max(dx, dy), dx, dy, basis='X', noise=biased_erasure_noise, p2q=0.1, bias=0.5, erasure_weight=0.2, biased_erasure=False, bias_preserving=True)
    # circuit = xzzx_circuit(max(dx, dy), dx, dy, basis='Z', noise=no_noise)
    # dem = circuit.detector_error_model(allow_gauge_detectors=True,decompose_errors=True, approximate_disjoint_errors=True)
    
    
    circuit = xzzx_circuit(cycles=max(dx,dy), dx=dx, dy=dy, basis='Z', noise=biased_erasure_noise,
                                p2q=0.1, bias=0, erasure_weight=1, biased_erasure=False,
                                bias_preserving=False)
    print(circuit)
    dem = circuit.detector_error_model(allow_gauge_detectors=True,decompose_errors=True, approximate_disjoint_errors=True, ignore_decomposition_failures=True)
    #circuit.diagram("timeline-svg")
    # graph = pymatching.Matching.from_detector_error_model(dem)
    # graph = pymatching.Matching.to_networkx(graph)
    # draw_graph(graph)
    #print(dem)
    print(repr(dem))
    #dem.diagram("matchgraph-3d")
    
    # circuit = xzzx_circuit(max(dx, dy), dx, dy, noise=two_q_gate_err, p2q=0.2)
    # circuit = biased_erasure_noise(circuit, p2q=0.01, bias=1, erasure_weight=0, biased_erasure=False, bias_preserving=True)
    


import stim
import numpy as np
from BiasedErasure.main_code.noise_channels import no_noise
from BiasedErasure.main_code.LogicalCircuit import LogicalCircuit
from typing import Union, List, Optional

"""
This code only works for logical X.
TODO: logical Z.
"""

def cluster_data_ids(dx: int, dy: int, cycles:int):
    # this function returns all the indexes of the data qubits in the cluster
    # num_layers = max(dx, dy)
    num_layers = 2*cycles - 1
    data_ids_list = []
    for layer_ix in range(num_layers):
        offset = layer_offset(dx, dy, layer_ix)
        for j in range(2*dy-1):
            for i in range(2*dx-1):
                if (i + j) % 2 == 0: #data qubit
                    qubit_ix = j*(2*dx-1)+i+offset
                    data_ids_list.append(qubit_ix)
                    # print(f"{i,j,layer_ix} qubit index is: {qubit_ix}. Now our list has: {data_ids_list}")
                        
    return data_ids_list


def cluster_qubits_ids(dx: int, dy: int, num_layers:int):
    # this function returns all the indexes of the qubits in the cluster
    # num_layers = max(dx, dy)
    qubits_ids_list = []
    for layer_ix in range(num_layers):
        offset = layer_offset(dx, dy, layer_ix)
        for j in range(2*dy-1):
            for i in range(2*dx-1):
                if (i + j) % 2 == 0: #data qubit
                    qubit_ix = j*(2*dx-1)+i+offset
                    qubits_ids_list.append(qubit_ix)
                elif (layer_ix % 2 == 0 and i % 2 == 0) or (layer_ix % 2 == 1 and i % 2 == 1): # ancilla qubits
                        qubit_ix = j*(2*dx-1)+i+offset
                        qubits_ids_list.append(qubit_ix)
                        
    return qubits_ids_list


def get_qubit_type(i: int, layer_ix: int):
    # this function gives us the data qubit type: X or Z type.
    if (layer_ix % 2 == 0 and i % 2 == 1) or (layer_ix % 2 == 1 and i % 2 == 0):
        return "X"
    elif (layer_ix % 2 == 0 and i % 2 == 0) or (layer_ix % 2 == 1 and i % 2 == 1):
        return "Z"
    
        

def layer_offset(dx: int, dy: int, layer_ix: int, physical=False) -> int:
    # if physical == True: only the number of physical qubits in the layer (half of the ancillas)
    # else: return the number full number of ancillas + data qubits for the regular circuit-based surface code
    n = int(np.ceil(0.5*(2*dx-1)*(2*dy-1)))  # num data qubit in a layer
    if physical:
        m = int(0.5*(2*dx-1)*(2*dy-1)*0.5)  # num ancilla qubits in a layer
    else:
        m = int(0.5*(2*dx-1)*(2*dy-1))  # num ancilla qubits in a surface code
    return layer_ix * (n + m)

def total_num_qubits(dx: int, dy: int, num_layers:int):
    n = int(np.ceil(0.5*(2*dx-1)*(2*dy-1)))  # num data qubits in each layer
    m = int(0.5*(2*dx-1)*(2*dy-1)*0.5) # num ancilla qubits in each layer
    # num_layers = max(dx,dy)
    return (n+m)*num_layers # num qubits in all layer

def qubit_index(dx: int, dy: int, row_index:int, col_index:int, layer_ix: int) -> int:
    outLayers_offset = layer_offset(dx=dx, dy=dy, layer_ix=layer_ix, physical=True)
    inLayer_offset = 0
    for j in range(2*dy-1):
        for i in range(2*dx-1):
            if row_index==i and col_index==j:
                return inLayer_offset+outLayers_offset
            if ((i+j)%2 == 0) or (layer_ix % 2 == 0 and i % 2 == 0) or (layer_ix % 2 == 1 and i % 2 == 1):
                inLayer_offset += 1

#####################################


def gates_in_layer(circ: stim.Circuit, dx: int, dy: int, bias_preserving_gates : bool, layer_ix=0, move_duration: Optional[float] = 200):
    # this function should change for different codes
    offset = layer_offset(dx, dy, layer_ix)
    data_qubit_x_type = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0 and ((i%2==0 and layer_ix%2==0) or (i%2==1 and layer_ix%2==1)))] )
    data_qubit_z_type = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0 and ((i%2==0 and layer_ix%2==1) or (i%2==1 and layer_ix%2==0)))])
    ancilla_qubit_even_layers = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1 and i%2 == 0)])
    ancilla_qubit_odd_layers = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1 and i%2 == 1)])
    
    
    # Entangling gates
    rnd1, rnd2, rnd3, rnd4 = [], [], [], []
    for j in range(2*dy-1):
        for i in range(2*dx-1):
            if (i + j) % 2 == 1:  # Iterate over ancilla qubits.
                if (layer_ix%2 == 0 and i%2 == 0) or (layer_ix%2 == 1 and i%2 == 1): # ancilla locations
                    if j != 2 * dy - 2: # north
                        rnd1.extend([j*(2*dx-1)+i+offset, (j+1)*(2*dx-1)+i+offset])
                    if i != 2 * dx - 2: # east
                        rnd2.extend([j*(2*dx-1)+i+offset, j*(2*dx-1)+(i+1)+offset])
                    if i != 0: # west
                        rnd3.extend([j*(2*dx-1)+i+offset, j*(2*dx-1)+(i - 1)+offset])
                    if j != 0: # south
                        rnd4.extend([j*(2*dx-1)+i+offset,  (j-1)*(2*dx-1)+i+offset])
    # round 1 - CZ: (no movement errors in first round, because we counted to movement when we moved the qubits to entangling zone)
    circ.append('CZ', rnd1)
    # circ.append('TICK')
    
    # round 2 - CX:
    if bias_preserving_gates:
        circ.append('CX', rnd2)
    else:
        targets = rnd2[1::2]
        circ.append('H', targets)
        circ.append('CZ', rnd2)
        circ.append('H', targets)
    circ.append('MOVE', (), move_duration) # idle errors during movement
    # circ.append('TICK')
    
    # round 3 - CX:
    if bias_preserving_gates:
        circ.append('CX', rnd3)
    else:
        targets = rnd3[1::2]
        circ.append('H', targets)
        circ.append('CZ', rnd3)
        circ.append('H', targets)
    circ.append('MOVE', (), move_duration) # idle errors during movement
    # circ.append('TICK')
    
    # round 4 - CZ:
    circ.append('CZ', rnd4)
    circ.append('MOVE', (), move_duration) # idle errors during movement
    # circ.append('TICK')

def interact_and_measure_all_layers(circ: stim.Circuit, dx: int, dy: int, num_layers: int, bias_preserving_gates : bool, basis='Z', move_duration: Optional[float] = 200):
    # In this function we take each layer from second to last, interact with previous layer (gates between layers) and measure previous layer.
    
    # take layer0 to the entangling zone:
    all_qubits_in_layer0 = sorted( [ j*(2*dx-1)+i for j in range(2*dy-1) for i in range(2*dx-1) if ( (i+j)%2 == 0 ) or ( (i+j)%2 == 1 and i%2==0 ) ] )
    circ.append('MOVE_TO_ENTANGLING', all_qubits_in_layer0, 0) # no idling errors because we can assume that we didn't remove this layer from the zone after preparation
    for layer_ix in range(num_layers):
        if layer_ix<num_layers-1:# last layer has no gates to the next layer
            layer2_ix = layer_ix + 1
            # TODO: take layer2 out of storage, entangle layer1 and measure layer1. make sure the measurements are at the same order for the detectors.
            offset = layer_offset(dx, dy, layer2_ix)
            all_qubits_in_layer2 = sorted( [ j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) if \
                                    ( (i+j)%2 == 0 ) or \
                                    ( (i+j)%2 == 1 and ((i%2==1 and layer2_ix%2==1) or (i%2==0 and layer2_ix%2==0)) ) ] )
            circ.append('MOVE_TO_ENTANGLING', all_qubits_in_layer2, 0)
            circ.append('MOVE', all_qubits_in_layer2, move_duration) # idle errors during movement
            gates_between_layers(circ=circ, dx=dx, dy=dy, layer1_ix=layer_ix, bias_preserving_gates=bias_preserving_gates)
            measure_layer(circ, dx, dy, layer_ix, num_layers=num_layers)
        elif layer_ix == num_layers-1: # last layer
            measure_layer(circ, dx, dy, layer_ix, num_layers=num_layers)


def gates_between_layers(circ: stim.Circuit, dx: int, dy: int, layer1_ix: int, bias_preserving_gates:bool):
    # TODO: take the relevant qubits outside of storage into the entangling zone.
    layer2_ix = layer1_ix + 1
    offset1, offset2 = layer_offset(dx, dy, layer1_ix), layer_offset(dx, dy, layer2_ix)

    # take layer2 out of storage (layer1 should already be inside the entangling zone)
    rnd = []
    for j in range(2*dy-1):
        for i in range(2*dx-1):
            if (i + j) % 2 == 0:  # Iterate over data qubits.
                if (layer1_ix % 2 == 0 and i % 2 == 0) or (layer1_ix % 2 == 1 and i % 2 == 1): # data 1: X type qubit
                    data1_ix = j*(2*dx-1)+i+offset1 # X type qubit
                    data2_ix = j*(2*dx-1)+i+offset2 # Z type qubit
                    rnd.extend([data1_ix, data2_ix]) # CX: X type to Z type
                elif (layer1_ix % 2 == 0 and i % 2 == 1) or (layer1_ix % 2 == 1 and i % 2 == 0): # data 1: Z type qubit
                    data1_ix = j*(2*dx-1)+i+offset1 # Z type qubit
                    data2_ix = j*(2*dx-1)+i+offset2 # X type qubit
                    rnd.extend([data2_ix, data1_ix]) # CX: X type to Z type
    if bias_preserving_gates:
        circ.append('CX', rnd)
    else:
        targets = rnd[1::2]
        circ.append('H', targets)
        circ.append('CZ', rnd)
        circ.append('H', targets)
    circ.append('TICK')


def measure_layer(circ, dx: int, dy: int, layer_ix: int, num_layers:int, move_duration: Optional[float] = 200):
    # Masure all qubits in the layer
    # We can only measure in Z and do H before if needed MX.
    offset = layer_offset(dx, dy, layer_ix)
    
    data_qubit_x_type = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0 and ((i%2==0 and layer_ix%2==0) or (i%2==1 and layer_ix%2==1)))] )
    data_qubit_z_type = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0 and ((i%2==0 and layer_ix%2==1) or (i%2==1 and layer_ix%2==0)))])
    ancilla_qubits = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1 and ((i%2==0 and layer_ix%2==0) or (i%2==1 and layer_ix%2==1)))] ) # all ancilla qubits are x-type

    circ.append('H', sorted(data_qubit_x_type + ancilla_qubits))
    if layer_ix != num_layers-1:
        circ.append('MOVE', (), move_duration) # Put idling noise on all active qubits due to movement time. last layer is measured in the entangling zone so there is no extra idling noise.
    circ.append('MZ', sorted(data_qubit_x_type + data_qubit_z_type + ancilla_qubits))
    circ.append('MOVE_TO_NO_NOISE', data_qubit_x_type + data_qubit_z_type + ancilla_qubits, 0) # all measured qubits are inactive now
        

def construct_detectors_all_layers(circ, dx: int, dy: int, num_layers:int, basis='Z'):
    # measure all the qubits in the cluster in the X basis (X type) or Z basis (Z type)
    # create 6 body operators parity checks for all the unit cells in the cluster
    # TODO: in XZZX cluster [], X type qubits are measured in X, Z type qubits are measured in Z.

    # num_layers = max(dx, dy)
    N = total_num_qubits(dx=dx, dy=dy, num_layers=num_layers) # total num qubits in all layers

    # step 1: measure all the qubits in the layer
    # for layer_ix in range(num_layers):
    #     measure_layer(circ, dx, dy, layer_ix)

    # step 2: for each qubit, check the 6 body operator where this qubit is the upper face qubit of the relevant unit cell:
    for layer_ix in range(num_layers):
        for j in range(2*dy-1):
            for i in range(2*dx-1):
                if (i + j) % 2 == 1:
                    if (layer_ix % 2 == 0 and i % 2 == 1) or (layer_ix % 2 == 1 and i % 2 == 0):  # (even) or (odd) layer
                        detector_targets = []
                        # offset = layer_offset(dx=dx, dy=dy, layer_ix=layer_ix)
                        
                        if j != 2 * dy - 2: # north
                            #data_ix = int(((j+1)*(2*dx-1)+i+offset)/2)  
                            qubit_ix = qubit_index(dx=dx, dy=dy, row_index=i, col_index=j+1, layer_ix=layer_ix)
                            detector_targets.append(stim.target_rec(-(N-qubit_ix)))
                        if i != 2 * dx - 2:  # east
                            # data_ix = int((j*(2*dx-1)+(i+1))/2)
                            qubit_ix = qubit_index(dx=dx, dy=dy, row_index=i+1, col_index=j, layer_ix=layer_ix)
                            detector_targets.append(stim.target_rec(-(N-qubit_ix)))
                        if i != 0:  # west
                            # data_ix = int((j*(2*dx-1)+(i - 1))/2)
                            qubit_ix = qubit_index(dx=dx, dy=dy, row_index=i-1, col_index=j, layer_ix=layer_ix)
                            detector_targets.append(stim.target_rec(-(N-qubit_ix)))
                        if j != 0:  # south
                            # data_ix = int(((j-1)*(2*dx-1)+i)/2)
                            qubit_ix = qubit_index(dx=dx, dy=dy, row_index=i, col_index=j-1, layer_ix=layer_ix)
                            detector_targets.append(stim.target_rec(-(N-qubit_ix)))
                        if layer_ix != num_layers - 1: # up
                            # data_ix = j*(2*dx-1)+i+layer_offset(dx=dx, dy=dy, layer_ix=layer_ix+1)
                            qubit_ix = qubit_index(dx=dx, dy=dy, row_index=i, col_index=j, layer_ix=layer_ix+1)
                            detector_targets.append(stim.target_rec(-(N-qubit_ix)))
                        if layer_ix != 0: # down
                            # data_ix = j*(2*dx-1)+i+layer_offset(dx=dx, dy=dy, layer_ix=layer_ix-1)
                            qubit_ix = qubit_index(dx=dx, dy=dy, row_index=i, col_index=j, layer_ix=layer_ix-1)
                            detector_targets.append(stim.target_rec(-(N-qubit_ix)))
                        circ.append('DETECTOR', detector_targets)



def initialize_layer(circ: stim.Circuit, dx: int, dy: int, basis='Z', layer_ix=0, move_duration: Optional[float] = 200):
    offset = layer_offset(dx, dy, layer_ix)
    # TODO: make this compatible with Z logical basis as well.
    ValueError(basis != 'X')

    data_qubit_x_type = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0 and ((i%2==0 and layer_ix%2==0) or (i%2==1 and layer_ix%2==1)))] )
    data_qubit_z_type = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 0 and ((i%2==0 and layer_ix%2==1) or (i%2==1 and layer_ix%2==0)))])
    ancilla_qubit_even_layers = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1 and i%2 == 0)])
    ancilla_qubit_odd_layers = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) if ((i+j)%2 == 1 and i%2 == 1)])
    
    # We can initialize only in Z and do H afterwards. also remove zones.
    circ.append('MOVE_TO_ENTANGLING', data_qubit_x_type + data_qubit_z_type, move_duration) # get idling noise
    circ.append('RZ', data_qubit_x_type + data_qubit_z_type) # X type data
    
    # Initialize data qubits
    if basis == 'Z':
        circ.append('H', data_qubit_z_type) # Z type in |+>
    elif basis == 'X':
        circ.append('H', data_qubit_x_type) # X type in |+>

    # Initialize ancilla qubits
    if layer_ix%2 == 0: # even layer
        circ.append('MOVE_TO_ENTANGLING', ancilla_qubit_even_layers, move_duration) # get idling noise
        circ.append('RZ', ancilla_qubit_even_layers)
        circ.append('H', ancilla_qubit_even_layers)
    else:
        circ.append('MOVE_TO_ENTANGLING', ancilla_qubit_odd_layers, move_duration) # get idling noise
        circ.append('RZ', ancilla_qubit_odd_layers)
        circ.append('H', ancilla_qubit_odd_layers)
    circ.append('TICK')
    
    
def initialize_all_layers(circ: stim.Circuit, dx: int, dy: int, num_layers: int, bias_preserving_gates : bool, basis='Z', move_duration: Optional[float] = 200):
    ValueError(basis != 'X')
    
    # initialize and connect all the qubits in all d layers
    
    ### part 1 - initialize all layers in 1 shot: (in XZZX cluster, X type qubits are initialize in |+>, Z type qubits in |0>)
    
    # We can entangle up to 4 layers in one shot, so we have batches of up to 4:
    for batch_start in range(0, num_layers, 4):
        batch_end = min(batch_start + 4, num_layers)
        offsets = []
        even_layers_offsets = []
        odd_layers_offsets = []
        # Inner loop to process each layer in the current batch
        for layer_ix in range(batch_start, batch_end):
            offset = layer_offset(dx, dy, layer_ix)
            offsets.append(offset)
            if layer_ix%2 == 0: # even layer
                even_layers_offsets.append(offset)
            elif layer_ix%2 == 1: # odd layer
                odd_layers_offsets.append(offset)
                    
        data_qubits_x_type_all_even_layers = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) for offset in even_layers_offsets if ( (i+j)%2 == 0 and (i%2==0) )] )
        data_qubits_x_type_all_odd_layers = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) for offset in odd_layers_offsets if  ( (i+j)%2 == 0 and (i%2==1) )] )
        data_qubits_x_type_all_layers = sorted(data_qubits_x_type_all_even_layers + data_qubits_x_type_all_odd_layers)
        
        data_qubits_z_type_all_even_layers = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) for offset in even_layers_offsets if ( (i+j)%2 == 0 and  i%2==1 )])
        data_qubits_z_type_all_odd_layers = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) for offset in odd_layers_offsets if ( (i+j)%2 == 0 and  i%2==0 )])
        data_qubits_z_type_all_layers = sorted(data_qubits_z_type_all_even_layers + data_qubits_z_type_all_odd_layers)

        ancilla_qubit_all_even_layers = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) for offset in even_layers_offsets if ((i+j)%2 == 1 and i%2 == 0)])
        ancilla_qubit_all_odd_layers = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) for offset in odd_layers_offsets if ((i+j)%2 == 1 and i%2 == 1)])
        
        # Initialize data qubits
        circ.append('MOVE_TO_ENTANGLING', data_qubits_x_type_all_layers + data_qubits_z_type_all_layers + ancilla_qubit_all_even_layers + ancilla_qubit_all_odd_layers, 0) # move all qubits to the entangling zone
        circ.append('RZ', sorted(data_qubits_x_type_all_layers + data_qubits_z_type_all_layers + ancilla_qubit_all_even_layers + ancilla_qubit_all_odd_layers)) # X type data  + all ancilla qubits
        circ.append('MOVE', data_qubits_x_type_all_layers + data_qubits_z_type_all_layers + ancilla_qubit_all_even_layers + ancilla_qubit_all_odd_layers, move_duration) # idling errors
        if basis == 'Z':
            circ.append('H', sorted(data_qubits_z_type_all_layers + ancilla_qubit_all_even_layers + ancilla_qubit_all_odd_layers)) # Z type in |+>
        elif basis == 'X':
            circ.append('H', sorted(data_qubits_x_type_all_layers + ancilla_qubit_all_even_layers + ancilla_qubit_all_odd_layers)) # X type in |+>

        circ.append('TICK')
        
        
        ### part 2 - gates in all layers in the batch in 1 shot:
        rnd1, rnd2, rnd3, rnd4 = [], [], [], []
        for layer_ix in range(batch_start, batch_end):
            offset = layer_offset(dx, dy, layer_ix)
            for j in range(2*dy-1):
                for i in range(2*dx-1):
                    if (i + j) % 2 == 1:  # Iterate over ancilla qubits.
                        if (layer_ix%2 == 0 and i%2 == 0) or (layer_ix%2 == 1 and i%2 == 1): # ancilla locations
                            if j != 2 * dy - 2: # north
                                rnd1.extend([j*(2*dx-1)+i+offset, (j+1)*(2*dx-1)+i+offset])
                            if i != 2 * dx - 2: # east
                                rnd2.extend([j*(2*dx-1)+i+offset, j*(2*dx-1)+(i+1)+offset])
                            if i != 0: # west
                                rnd3.extend([j*(2*dx-1)+i+offset, j*(2*dx-1)+(i - 1)+offset])
                            if j != 0: # south
                                rnd4.extend([j*(2*dx-1)+i+offset,  (j-1)*(2*dx-1)+i+offset])
        
        # round 1 - CZ: (no movement errors in first round, because we counted to movement when we moved the qubits to entangling zone)
        circ.append('CZ', rnd1)
        
        # round 2 - CX:
        if bias_preserving_gates:
            circ.append('CX', rnd2)
        else:
            targets = rnd2[1::2]
            circ.append('H', targets)
            circ.append('CZ', rnd2)
            circ.append('H', targets)
        circ.append('MOVE', (), move_duration) # idle errors during movement
        
        # round 3 - CX:
        if bias_preserving_gates:
            circ.append('CX', rnd3)
        else:
            targets = rnd3[1::2]
            circ.append('H', targets)
            circ.append('CZ', rnd3)
            circ.append('H', targets)
        circ.append('MOVE', (), move_duration) # idle errors during movement
        
        # round 4 - CZ:
        circ.append('CZ', rnd4)
        circ.append('MOVE', (), move_duration) # idle errors during movement
        
        # move all qubits to the storage zone:
        circ.append('MOVE_TO_STORAGE', ancilla_qubit_all_even_layers + ancilla_qubit_all_odd_layers + data_qubits_x_type_all_layers + data_qubits_z_type_all_layers, move_duration) # idle errors during movement



        # # move all qubits (except first layer) to the storage zone:
        # data_qubits_x_type_all_layers_except_0 = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) for offset in offsets if ( (i+j)%2 == 0 and (i%2==0) and offset > 0)] )
        # data_qubits_z_type_all_even_layers_except_0 = sorted([j*(2*dx-1)+i+offset for j in range(2*dy-1) for i in range(2*dx-1) for offset in offsets if ( (i+j)%2 == 0 and  i%2==1 and offset > 0)])

        # circ.append('MOVE_TO_STORAGE', ancilla_qubit_all_even_layers + ancilla_qubit_all_odd_layers + data_qubits_x_type_all_layers_except_0 + data_qubits_z_type_all_even_layers_except_0, move_duration) # idle errors during movement

    

def logical(circ: stim.Circuit, dx: int, dy: int, num_layers:int, basis='Z') -> None:
    # Extract the logical qubit information from the data qubits in last layer
    # This function should change for different codes
    # num_layers = max(dx, dy)
    N = total_num_qubits(dx=dx, dy=dy, num_layers=num_layers) # total num qubits in all layer
    detector_targets = []
    # offset = layer_offset(dx=dx, dy=dy, layer_ix=max(dx,dy)-1, physical=True)
    if basis == 'X':
        for layer_ix in range(num_layers):
            for j in range(2*dy-1):
                if layer_ix % 2 == 0 and j % 2 == 0: # left col on even layers
                    data_ix = qubit_index(dx=dx, dy=dy, row_index=0, col_index=j, layer_ix=layer_ix)
                    # print(f"i= {0}, j={j}, data index = {data_ix}")
                    detector_targets.append(stim.target_rec(-(N-data_ix)))
        if num_layers % 2 == 0:  # last layer measure in X if even num of layers (d is even)
            for j in range(2*dy-1):
                data_ix = qubit_index(dx=dx, dy=dy, row_index=0, col_index=j, layer_ix=num_layers-1)
                detector_targets.append(stim.target_rec(-(N-data_ix)))
                # print(f"Last layer! i= {0}, j={j}, data index = {num_layers-1}")
        circ.append('OBSERVABLE_INCLUDE', detector_targets, 0)
    else:
        assert basis == 'Z'
        print(f"The circuit is not compatible with Z basis initialization.")
        for layer_ix in range(num_layers):
            for i in range(2*dx-1): # first row on even layers
                if layer_ix % 2 == 0 and i % 2 == 0:
                    data_ix = qubit_index(dx=dx, dy=dy, row_index=i, col_index=0, layer_ix=layer_ix)
                    detector_targets.append(stim.target_rec(-(N-data_ix)))
        if num_layers % 2 == 1:  # last layer measure in Z if odd num of layers (d is even)
            for i in range(2*dx-1):
                data_ix = qubit_index(dx=dx, dy=dy, row_index=i, col_index=0, layer_ix=num_layers-1)
                detector_targets.append(stim.target_rec(-(N-data_ix)))
        circ.append('OBSERVABLE_INCLUDE', detector_targets, 0)
        
    
def xzzx_cluster_noisy(cycles, dx: int, dy: int, bias_preserving_gates : bool, basis='Z', noise=no_noise, **kwargs):
    num_layers = 2*cycles - 1 # every layer checks X or Z, and we want cycles XZ layers in total. plus 1 because of the init layer
    error_model = {
        "entangling_zone_error_rate": (.002 / 4, .002 / 4, .005 / 4),
        "entangling_gate_error_rate": (.002 / 4, .002 / 4, .0025 / 4, .002 / 4, 0, 0, 0, .002 / 4, 0, 0, 0, .0025 / 4, 0, 0, .005 / 4),
        "idle_loss_rate": 1e-7,
        "idle_error_rate": (5e-6 / 25, 5e-6 / 25, 2e-5 / 25),
        "entangling_loss_rate": 0.005/4,
        "reset_error_rate": 0.003,
        "measurement_error_rate": 0.004,
        "single_qubit_error_rate": (1e-4, 1e-4, 1e-4),
        "reset_loss_rate": 0.000
    }
    
    move_duration = 200
    qubit_indices = cluster_qubits_ids(dx,dy, num_layers=num_layers)
    circ = LogicalCircuit(qubit_indices = qubit_indices, initialize_circuit=False, **error_model)
        
    initialize_all_layers(circ, dx, dy, num_layers=num_layers, basis=basis, bias_preserving_gates=bias_preserving_gates)
    interact_and_measure_all_layers(circ, dx, dy, num_layers=num_layers, basis=basis, bias_preserving_gates=bias_preserving_gates)
    # all_gates(circ, dx, dy, num_layers=num_layers, basis=basis)
    construct_detectors_all_layers(circ, dx, dy, num_layers=num_layers, basis=basis)
    # circ = noise(circ, **kwargs)
    logical(circ, dx, dy, basis=basis, num_layers=num_layers)
    return circ

def get_neighbors_next_previous_layers(dx: int, dy: int, qubit_index):
    n = int(np.ceil(0.5*(2*dx-1)*(2*dy-1)))  # num data qubit in a layer
    m = int((2*dx-1)*(2*dy-1)*0.5)  # num ancilla qubits in a surface code (twice num ancilla in MBQC layer)
    offset = n + m
    neighbors_indices = []
    if qubit_index - offset >= 0:
        neighbors_indices.append(qubit_index - offset)
    if qubit_index + offset < max(dx, dy) * offset - 1:
        neighbors_indices.append(qubit_index + offset)
    return neighbors_indices
        

if __name__ == '__main__':
    dx = 2
    dy = 2
    circuit = xzzx_cluster(cycles=max(dx,dy), dx=dx, dy=dy, basis='Z', noise=no_noise)
    print(circuit)


import networkx as nx
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.cbook as cbook
from matplotlib import cm
from matplotlib.ticker import NullFormatter
import numpy as np
from scipy.optimize import curve_fit
import copy

def draw_graph2(g, save=False, filename=None, from_edges=False, size = (2,2)):
    if from_edges:
        g_ = nx.Graph()
        n_nodes = max([m for e in g for m in e])
        g_.add_nodes_from(list(range(n_nodes)))
        g_.add_edges_from(g)
        g = g_
    colour_map = []
    for n in range(len(g.nodes)):
        colour_map.append('gold')
    pos = nx.kamada_kawai_layout(g)
    plt.figure(figsize=size)
    nx.draw(g, pos=pos, node_color=colour_map, with_labels=True)

    edge_labels = nx.get_edge_attributes(g, "weight")
    print([e for e in edge_labels])
    #print(pos)
    nx.draw_networkx_edge_labels(g, pos, edge_labels)
    if save:
        plt.savefig(filename)
    plt.show()

def draw_graph(G):

    elarge = [(u, v) for (u, v, d) in G.edges(data=True) if d["weight"] > 1]
    esmall = [(u, v) for (u, v, d) in G.edges(data=True) if d["weight"] <= 1]

    pos = nx.spring_layout(G, seed=7)  # positions for all nodes - seed for reproducibility

    # nodes
    nx.draw_networkx_nodes(G, pos, node_size=700)

    # edges
    nx.draw_networkx_edges(G, pos, edgelist=elarge, width=6)
    nx.draw_networkx_edges(
        G, pos, edgelist=esmall, width=6, alpha=0.5, edge_color="b", style="dashed"
    )

    # node labels
    nx.draw_networkx_labels(G, pos, font_size=20, font_family="sans-serif")
    # edge weight labels
    edge_labels = nx.get_edge_attributes(G, "weight")

    for key, value in edge_labels.items():
        edge_labels[key] = round(value,2)
    nx.draw_networkx_edge_labels(G, pos, edge_labels)

    ax = plt.gca()
    ax.margins(0.08)
    plt.axis("off")
    plt.tight_layout()
    plt.show()
    return edge_labels



def model_func(x, A, B, C, threshold, alpha, beta):
    p_physical = x[0]
    n = x[1]
    x_val = (p_physical - threshold) * alpha * n**beta
    p_logical = A + B * x_val + C * x_val**2
    return p_logical

def fit_model(p_physical, p_logical, n):
    def fit_func(x, A, B, C, threshold, alpha, beta):
        return model_func(x, A, B, C, threshold, alpha, beta) - p_logical

    popt, _ = curve_fit(fit_func, (p_physical, n), p_logical, method='lm')

    A_fit, B_fit, C_fit, threshold_fit, alpha_fit, beta_fit = popt
    return alpha_fit, beta_fit, threshold_fit, A_fit, B_fit, C_fit



# Functions for fitting the threshold
def CriticalExponentFit(xdata_tuple, pc, nu, A, B, C):
    p, d = xdata_tuple
    x = (p - pc)*d**(1/nu)
    pl = A + B*x + C*x**2
    return pl

def EmpericalFit(xdata_tuple, pc, A):
    p, d = xdata_tuple
    pl = A*(p/pc)**(d/2)
    return pl

def FitDistance(p, A, d):
    pl = A*p**(d/2)
    return pl

def DistanceEst(sweep_p_list, sweep_pl_total_list, if_plot=False):
    num_p = len(sweep_p_list)
    num_code = len(sweep_pl_total_list)
    sweep_d_list = []
    for sweep_pl_list in sweep_pl_total_list:
        initial_guess = (0.01, 3)
        popt, pcov = curve_fit(FitDistance, np.array(sweep_p_list), np.array(sweep_pl_list) + 1e-10, p0=initial_guess)
        sweep_d_list.append(popt[1])

    return sweep_d_list

 
       
def ThresholdEst_extrapolation(sweep_p_list, sweep_pl_total_list, if_plot=False):
    num_p = len(sweep_p_list)
    num_code = len(sweep_pl_total_list)
    sweep_d_list = DistanceEst(sweep_p_list, sweep_pl_total_list, if_plot=False)
    
    fit_d_list = copy.deepcopy(sweep_d_list)
    sweep_p_list = list(sweep_p_list)*num_code
    sweep_d1_list = []
    for sweep_d in sweep_d_list:
        sweep_d1_list += [sweep_d]*num_p
    sweep_d_list = sweep_d1_list
    sweep_pl_list = list(np.reshape(np.array(sweep_pl_total_list) + 1e-10, [num_p*num_code, ]))
    
    fit_X = np.vstack([np.reshape(np.array(sweep_p_list), [1, num_p*num_code]), 
                       np.reshape(np.array(sweep_d_list), [1, num_p*num_code])])
    fit_Z = np.reshape(np.array(sweep_pl_total_list), [num_p*num_code, ])
    initial_guess = (0.04, 0.1)
    popt, pcov = curve_fit(EmpericalFit, fit_X, fit_Z, p0=initial_guess)
    perr = np.sqrt(np.diag(pcov))
    
    # plot
    p_c, A = popt[0], popt[1]
    fit_p_list = list(set(sweep_p_list))
    fit_pl_list = np.reshape(np.array(sweep_pl_list), [len(fit_d_list), len(fit_p_list)])
    if if_plot:
        fitted_pl_list = []
        for sweep_d in fit_d_list:
            fitted_pl_list.append([EmpericalFit((sweep_p, sweep_d), p_c, A) for sweep_p in fit_p_list])
        
        plt.figure()
        for i in range(len(fit_d_list)):
            plt.plot(fit_p_list, fitted_pl_list[i], '-', c = 'C%i'%i)
            plt.plot(sweep_p_list[:num_p], sweep_pl_list[i*num_p:(i + 1)*num_p], 'D', c = 'C%i'%i)
        plt.xscale('log')
        plt.yscale('log')
        plt.xlabel('p')
        plt.ylabel('WER')
    
    print('p_c:', popt[0])
    
    return popt[0]

def DistanceEst(sweep_p_list, sweep_pl_total_list, if_plot=False):
    num_p = len(sweep_p_list)
    num_code = len(sweep_pl_total_list)
    sweep_d_list = []
    for sweep_pl_list in sweep_pl_total_list:
        initial_guess = (0.01, 3)
        popt, pcov = curve_fit(FitDistance, np.array(sweep_p_list), np.array(sweep_pl_list) + 1e-10, p0=initial_guess)
        sweep_d_list.append(popt[1])

    return sweep_d_list

def FitDistance(p, A, d):
    pl = A*p**(d/2)
    return pl

 

if __name__ == '__main__':
    # Example usage
    p_physical = np.array([1.0, 2.0, 3.0, 4.0])
    p_logical = np.array([5.0, 8.0, 15.0, 26.0])
    n = 10  # Your choice of n

    alpha, beta, threshold, A, B, C = fit_model(p_physical, p_logical, n)
    print("alpha:", alpha)
    print("beta:", beta)
    print("threshold:", threshold)
    print("A:", A)
    print("B:", B)
    print("C:", C)
